"""
An example of running DEISM-ARG with a single parameter set,
the parameters are defined in the configSingleParam_ARG.yaml file.
You can also set the parameters and run the codes via the command line.
"""

import time
import os
import psutil
import numpy as np
import ray

from deism.core_deism import (
    pre_calc_Wigner,
    init_source_directivities_ARG,
    init_receiver_directivities_ARG,
    vectorize_C_vu_r,
    vectorize_C_nm_s_ARG,
    run_DEISM_ARG,
)

from deism.core_deism_arg import (
    Room_deism_cpp,
    get_ref_paths_ARG,
    rotate_room_src_rec,
    find_wall_centers,
)
from deism.data_loader import (
    cmdArgsToDict,
    printDict,
    detect_conflicts,
)
from deism.utilities import plot_RTFs


def init_parameters_convex(params):
    """
    Initialize some important additional parameters for DEISM-ARG
    In addition to the ones defined in configSingleParam_ARG.yaml
    """
    # Room vertices
    # A simple room with 8 vertices
    vertices = np.array(
        [
            [0, 0, 0],  # Origin, but it does not have to be the origin
            [0, 0, 3.5],  # [x, y, z] coordinates of the room vertices
            [0, 3, 2.5],
            [0, 3, 0],
            [4, 0, 0],
            [4, 0, 3.5],
            [4, 3, 2.5],
            [4, 3, 0],
        ]
    )
    # --- Room rotation, if rotate the room w.r.t the origin ---
    if_rotate_room = 1
    # --- Room rotation angles using Z-X-Z Euler angles ---
    # [alpha, beta, gamma], 3D Euler angles, The rotation matrix calculation used in COMSOL, see:
    # https://doc.comsol.com/5.5/doc/com.comsol.help.comsol/comsol_ref_definitions.12.092.html
    # The chosen rotation angles are used to illustrate a scenario where we can compare with the room created using pyroomacoustics
    # See file DEISM_ARG_comparisons.py
    room_rotation = np.array([90, 90, 90])  # [alpha, beta, gamma] in degrees
    # --- Add the above parameters to the params dictionary ---
    params["vertices"] = vertices
    params["wallCenters"] = find_wall_centers(vertices)
    params["if_rotate_room"] = if_rotate_room
    params["room_rotation"] = room_rotation
    # Apply room rotation to the room vertices and source/receiver positions
    if if_rotate_room:
        params = rotate_room_src_rec(params)
    return params


def main():
    # Load the parameters from the configSingleParam_ARG.yaml file amd command line
    params, cmdArgs = cmdArgsToDict(mode="RTF", roomtype="convex")
    # Initialize some additional parameters for DEISM-ARG
    # For example, room vertices, room rotation, etc.
    params = init_parameters_convex(params)
    # detect conflicts
    detect_conflicts(params)
    # print the parameters or not
    if cmdArgs.quiet:
        params["silentMode"] = 1
    printDict(params)
    # Save only the current parameters
    params_save = params.copy()

    # If run DEISM function, run if --run flag is set in the cmd
    # If cmdArgs are all None values, run following codes directily
    if cmdArgs.run or all(
        value in [None, False] for value in vars(cmdArgs).values()
    ):  # no input in cmd will also run
        # Initialize directivities
        # -------------------------------------------------------
        # Shared parameters for all modes, ORG, LC, MIX
        # -------------------------------------------------------
        # params = init_receiver_directivities_ARG(
        #     params, params["if_rotate_room"], room_rotation=params["room_rotation"]
        # )
        # Using DESIM-ARG with c++ libroom_deism to find the visible images and reflection matrices, attenuations
        # initialize Room_deism
        room_pra_deism = Room_deism_cpp(params)
        # Plot the room and the images
        room_pra_deism.plot_room()
        # get the reflection paths
        params = get_ref_paths_ARG(params, room_pra_deism)
        # Calculating the reflected source directivity coefficients
        params = init_source_directivities_ARG(
            params,
            params["if_rotate_room"],
            params["reflection_matrix"],
            room_rotation=params["room_rotation"],
        )
        # -------------------------------------------------------
        # Some additional parameters for different modes
        # -------------------------------------------------------
        # If we use LC or MIX mode, we need to vectorize the directivity coefficients
        if params["DEISM_mode"] == "LC" or params["DEISM_mode"] == "MIX":
            # Vectorize the directivity data, used for LC and MIX mode
            params = vectorize_C_vu_r(params)
            params = vectorize_C_nm_s_ARG(params)
        # -------------------------------------------------------
        # If we use the ORG and MIX mode, we need to precompute the Wigner 3J matrices
        if params["DEISM_mode"] == "ORG" or params["DEISM_mode"] == "MIX":
            params["Wigner"] = pre_calc_Wigner(params)
        # -------------------------------------------------------
        # Run DEISM-ARG parallel processing using Ray
        # -------------------------------------------------------
        # Initialize Ray
        num_cpus = psutil.cpu_count(logical=False)
        if not ray.is_initialized():
            ray.init(num_cpus=num_cpus)
            print("\n")
        # Run DEISM-ARG for specified mode
        P_DEISM_ARG = run_DEISM_ARG(params)
        # Shutdown Ray
        ray.shutdown()
        # -------------------------------------------------------
        # Plot the RTFs and save the figure
        # -------------------------------------------------------
        # We now also save some parameters in the name of the figure
        figure_name = "DEISM_ARGs_RO_{}_{}_vertices_src_{:.2f}_{:.2f}_{:.2f}_rec_{:.2f}_{:.2f}_{:.2f}".format(
            params["maxReflOrder"],
            len(params["vertices"]),
            params["posSource"][0],
            params["posSource"][1],
            params["posSource"][2],
            params["posReceiver"][0],
            params["posReceiver"][1],
            params["posReceiver"][2],
        )
        save_path = "./outputs/figures"
        # check if the save path exists
        if not os.path.exists(save_path):
            os.makedirs(save_path)
        # Parameters for plotting the RTFs
        P_all = [P_DEISM_ARG]  # A list of RTFs
        P_labels = ["DEISM-ARG-{}".format(params["DEISM_mode"])]  # A list of labels
        P_freqs = [params["freqs"]]  # A list of frequency bins
        PLOT_SCALE = "dB"  # Plot scale, can be "dB" or "linear"
        IF_FREQS_DB = 1  # If the frequency bins are plotted in dB
        IF_SAME_MAGSCALE = 0  # If the magnitude scale is fixed the same for all RTFs
        IF_UNWRAP_PHASE = 0  # If the phase is unwrapped
        IF_SAVE_PLOT = 1  # If save the plot
        # Plot the RTFs
        plot_RTFs(
            figure_name,
            save_path,
            P_all,
            P_labels,
            P_freqs,
            PLOT_SCALE,
            IF_FREQS_DB,
            IF_SAME_MAGSCALE,
            IF_UNWRAP_PHASE,
            IF_SAVE_PLOT,
        )
        # -------------------------------------------------------
        # Save the results to local directory with .npz format
        # ------------------------------------------------
        save_path = "./outputs/RTFs"
        # check if the save path exists
        if not os.path.exists(save_path):
            os.makedirs(save_path)
        # Save the results along with all the parameters to a .npz file with file name as the current time
        # Save only the initial parameters which are used to run the DEISM function
        # You can change the naming to other formats if you want
        # We name the file with the current time for simplicity
        np.savez(
            f"{save_path}/DEISM_ARG_RTF_{time.strftime('%Y%m%d_%H%M%S')}",
            P_DEISM=P_DEISM_ARG,
            params=params_save,
        )
        # read the saved file
        # data = np.load(f"{save_path}/DEISM_ARG_RTF_.npz")
    else:
        print("DEISM Function has not been run because --run flag was not set.\n")


# -------------------------------------------------------
if __name__ == "__main__":
    main()



"""
Python wrapper for C++ reflection counting function.

Usage:
    from deism.count_reflections_wrapper import count_reflections_cpp

    count = count_reflections_cpp(order, room_dims, c, T60)
"""

import ctypes
import os
import platform


def _load_cpp_library():
    """Load the compiled C++ library."""
    # Determine the library extension based on platform
    if platform.system() == "Darwin":  # macOS
        lib_ext = ".dylib"
    elif platform.system() == "Linux":
        lib_ext = ".so"
    elif platform.system() == "Windows":
        lib_ext = ".dll"
    else:
        lib_ext = ".so"

    # Try to find the library in the same directory
    lib_path = os.path.join(os.path.dirname(__file__), f"count_reflections{lib_ext}")

    if not os.path.exists(lib_path):
        raise FileNotFoundError(
            f"C++ library not found at {lib_path}. "
            f"Please compile count_reflections.cpp first:\n"
            f"  cd {os.path.dirname(__file__)}\n"
            f"  g++ -shared -fPIC -O3 -std=c++11 count_reflections.cpp -o count_reflections{lib_ext}"
        )

    # Load the library
    lib = ctypes.CDLL(lib_path)

    # Define the function signature
    lib.count_reflections_shoebox_test.argtypes = [
        ctypes.c_int,  # order
        ctypes.c_double,  # Lx
        ctypes.c_double,  # Ly
        ctypes.c_double,  # Lz
        ctypes.c_double,  # c
        ctypes.c_double,  # T60
    ]
    lib.count_reflections_shoebox_test.restype = ctypes.c_longlong

    return lib


# Try to load the library (will fail if not compiled)
try:
    _cpp_lib = _load_cpp_library()
    _cpp_available = True
except (FileNotFoundError, OSError) as e:
    _cpp_available = False
    _cpp_error = str(e)


def count_reflections_cpp(order, room_dims, c, T60):
    """
    Fast C++ implementation of reflection path counting.

    Args:
        order: Maximum reflection order
        room_dims: Tuple or array of (Lx, Ly, Lz) room dimensions
        c: Speed of sound
        T60: Reverberation time

    Returns:
        int: Number of reflection paths

    Raises:
        RuntimeError: If C++ library is not compiled
    """
    if not _cpp_available:
        raise RuntimeError(
            f"C++ library not available. {_cpp_error}\n"
            "To compile:\n"
            "  cd deism\n"
            "  g++ -shared -fPIC -O3 -std=c++11 count_reflections.cpp -o count_reflections.dylib  # macOS\n"
            "  g++ -shared -fPIC -O3 -std=c++11 count_reflections.cpp -o count_reflections.so    # Linux\n"
            "  g++ -shared -fPIC -O3 -std=c++11 count_reflections.cpp -o count_reflections.dll   # Windows"
        )

    Lx, Ly, Lz = room_dims[0], room_dims[1], room_dims[2]

    count = _cpp_lib.count_reflections_shoebox_test(
        int(order), float(Lx), float(Ly), float(Lz), float(c), float(T60)
    )

    return count


if __name__ == "__main__":
    # Test the function
    import time
    import numpy as np

    room_dims = (3.29, 6.23, 2.58)
    c = 343
    T60 = 0.827
    order = 50

    print("Testing C++ reflection counting function...")
    print(f"Order: {order}, Room dims: {room_dims}, c: {c}, T60: {T60}")

    if _cpp_available:
        start = time.time()
        count = count_reflections_cpp(order, room_dims, c, T60)
        elapsed = time.time() - start
        print(f"Count: {count}")
        print(f"Time: {elapsed:.4f} seconds")
    else:
        print(f"Error: {_cpp_error}")



"""
Testing DEISM with single parameter settings
"""

# -------------------------------------------------------
# Authors: Zeyu Xu
# Songjiang Tan
# Email: zeyu.xu@audiolabs-erlangen.de
# -------------------------------------------------------
import yaml
import argparse
import os
import time
import numpy as np
import matplotlib.pyplot as plt
from scipy import special as scy
import ray
import psutil
from deism.core_deism import *


def main():
    params, cmdArgs = cmdArgsToDict()
    # print the parameters or not
    if cmdArgs.quiet:
        params["silentMode"] = 1
    printDict(params)
    # Save only the current parameters
    params_save = params.copy()
    # If run DEISM function, run if --run flag is set in the cmd
    # If cmdArgs are all None values, run following codes directily
    if cmdArgs.run or all(
        value in [None, False] for value in vars(cmdArgs).values()
    ):  # no input in cmd will also run
        # Detect conflicts between the parameters
        detect_conflicts(params)
        # Initialize directivities
        params = init_receiver_directivities(params)
        params = init_source_directivities(params)
        # Precompute reflection paths
        images = pre_calc_images_src_rec(params)
        # If use DEISM-ORG or DEISM-LC, merge images
        if params["DEISM_mode"] == "ORG" or params["DEISM_mode"] == "LC":
            images = merge_images(images)
        params["images"] = images
        # If use DEISM-LC or DEISM-MIX, vectorize the directivity coefficients
        if params["DEISM_mode"] == "LC" or params["DEISM_mode"] == "MIX":
            # Vectorize the directivity data, used for DEISM-LC
            params = vectorize_C_nm_s(params)
            params = vectorize_C_vu_r(params)
        # If use DEISM-ORG or DEISM-MIX, precompute Wigner 3J matrices
        if params["DEISM_mode"] == "ORG" or params["DEISM_mode"] == "MIX":
            params["Wigner"] = pre_calc_Wigner(params)
        # -------------------------------------------------------
        # Initialize Ray
        num_cpus = psutil.cpu_count(logical=False)
        if not ray.is_initialized():
            ray.init(num_cpus=num_cpus)
            print("\n")
        P = run_DEISM(params)
        # Shutdown Ray
        ray.shutdown()
        # -------------------------------------------------------
        # Save the results to local directory with .npz format
        save_path = "./outputs/RTFs"
        # check if the save path exists
        if not os.path.exists(save_path):
            os.makedirs(save_path)
        # Save the results along with all the parameters to a .npz file with file name as the current time
        np.savez(
            f"{save_path}/DEISM_shoebox_RTFs_{time.strftime('%Y%m%d_%H%M%S')}",
            P_DEISM=P,
            params=params_save,
        )

    else:
        print("DEISM Function has not been run because --run flag was not set.\n")


# -------------------------------------------------------
if __name__ == "__main__":
    main()



import fnmatch
import yaml
import argparse
import os
import scipy.io as sio
import numpy as np


class ConflictChecks:
    """Static methods for efficient parameter conflict checking"""

    @staticmethod
    def directivity_checks(params):
        """Check directivity-related parameter conflicts"""
        # If monopole is used, set spherical harmonic order to 0 for source and receiver
        if params["sourceType"] == "monopole":
            params["sourceOrder"] = 0
        if params["receiverType"] == "monopole":
            params["receiverOrder"] = 0
            params["ifReceiverNormalize"] = 0

        # If not monopole, raise a warning if the spherical harmonic order is also 0
        if params["sourceType"] != "monopole" and params["sourceOrder"] == 0:
            print(
                "[Warning] Spherical harmonic order is set to 0 for source, but source type is not monopole! \n"
            )
        if params["receiverType"] != "monopole" and params["receiverOrder"] == 0:
            print(
                "[Warning] Spherical harmonic order is set to 0 for receiver, but receiver type is not monopole! \n"
            )
        if params["receiverType"] != "monopole" and params["ifReceiverNormalize"] == 0:
            print(
                "[Warning] Receiver normalization is set to 0 for non-monopole receiver, make sure you know what you are doing! \n"
            )

    @staticmethod
    def distance_spheres_checks(params):
        """Check distance-related parameter conflicts"""
        # If either source or receiver is not monopole, check the distance between source and receiver
        if params["sourceType"] != "monopole":
            if params["radiusSource"] is None:
                raise ValueError("Source radius is not set for non-monopole source")
            # If distance between source and receiver is smaller than the source radius, raise a warning
            if (
                np.linalg.norm(params["posSource"] - params["posReceiver"])
                < params["radiusSource"]
            ):
                print(
                    "[Warning] Distance between source and receiver is smaller than the source radius! \n"
                )

        if params["receiverType"] != "monopole":
            if params["radiusReceiver"] is None:
                raise ValueError("Receiver radius is not set for non-monopole receiver")
            # If distance between source and receiver is smaller than the receiver radius, raise a warning
            if (
                np.linalg.norm(params["posSource"] - params["posReceiver"])
                < params["radiusReceiver"]
            ):
                print(
                    "[Warning] Distance between source and receiver is smaller than the receiver radius! \n"
                )

        # If both source and receiver are not monopole, check the distance between source and receiver
        if params["sourceType"] != "monopole" and params["receiverType"] != "monopole":
            if (
                np.linalg.norm(params["posSource"] - params["posReceiver"])
                < params["radiusSource"] + params["radiusReceiver"]
            ):
                print(
                    "[Warning] Distance between source and receiver is smaller than the sum of the source and receiver radius! \n"
                )

    @staticmethod
    def wall_material_checks(params):
        """Check wall material parameter conflicts"""
        # Check if multiple material types are defined
        if len(params.get("givenMaterials", [])) > 1:
            raise ValueError(
                "The user can only define one of the following three parameters: "
                "impedance, absorption coefficient, or reverberation time. "
                f"The following parameters are defined: {', '.join(params['givenMaterials'])}"
            )

        # Check the dimension of the impedance, absorption coefficient, and reverberation time
        # data shapes should be 2D numpy arrays for impedance, absorption coefficient, first dimension is 6
        # and 1D numpy arrays for reverberation time
        if params.get("givenMaterials") == "impedance":
            if "impedance" in params:
                arr = params["impedance"]
                if not (
                    isinstance(arr, np.ndarray) and arr.ndim == 2 and arr.shape[0] == 6
                ):
                    raise ValueError(
                        f"Impedance data must be a 2D numpy array with 6 rows (for shoebox room), "
                        f"got shape {arr.shape if isinstance(arr, np.ndarray) else 'not a numpy array'}"
                    )
        elif params.get("givenMaterials") == "absorpCoefficient":
            if "absorptionCoeff" in params:
                arr = params["absorptionCoeff"]
                if not (
                    isinstance(arr, np.ndarray) and arr.ndim == 2 and arr.shape[0] == 6
                ):
                    raise ValueError(
                        f"Absorption coefficient data must be a 2D numpy array with 6 rows (for shoebox room), "
                        f"got shape {arr.shape if isinstance(arr, np.ndarray) else 'not a numpy array'}"
                    )
        elif params.get("givenMaterials") == "reverberationTime":
            if "reverberationTime" in params:
                arr = params["reverberationTime"]
                if not (isinstance(arr, np.ndarray) and arr.ndim == 1):
                    raise ValueError(
                        f"Reverberation time data must be a 1D numpy array, "
                        f"got shape {arr.shape if isinstance(arr, np.ndarray) else 'not a numpy array'}"
                    )

    @staticmethod
    def distance_boundaries_checks(params):
        """Check distance from source and receiver to the boundaries of the room"""
        # The source and receiver should be inside the room, not touching the boundaries
        # If they are on the boundaries, raise a warning and stop the program
        # If they are too close to the boundaries, raise a message and continue the program
        # TODO
        if params["roomType"] == "shoebox":
            src_x = params["posSource"][0]
            src_y = params["posSource"][1]
            src_z = params["posSource"][2]
            rec_x = params["posReceiver"][0]
            rec_y = params["posReceiver"][1]
            rec_z = params["posReceiver"][2]
            room_x = params["roomSize"][0]
            room_y = params["roomSize"][1]
            room_z = params["roomSize"][2]
            if (
                src_x <= 0
                or src_x >= room_x
                or src_y <= 0
                or src_y >= room_y
                or src_z <= 0
                or src_z >= room_z
            ):
                print("[Warning] Source is on the boundaries or outside of the room!")
            if (
                rec_x <= 0
                or rec_x >= room_y
                or rec_y <= 0
                or rec_y >= room_y
                or rec_z <= 0
                or rec_z >= room_z
            ):
                print("[Warning] Receiver is on the boundaries or outside of the room!")

    @staticmethod
    def check_all_conflicts(params):
        """Run all conflict checks efficiently"""
        ConflictChecks.directivity_checks(params)
        ConflictChecks.distance_spheres_checks(params)
        ConflictChecks.distance_boundaries_checks(params)
        ConflictChecks.wall_material_checks(params)


def load_format_materials_checks(datain, datatype):
    """
    Format the materials to the correct shape
    inputs:
    - datain: Could be many types of inputs, including:
    - datatype: str, the type of the parameters to be converted
    Outputs:
    - if input is an integer or float number, return a array of size (6, 1) for impedance, absorption coefficients; return a array of size (1, 1) for reverberation time;
    - if input is a list of size 6, return a array of size (6, 1) for impedance, absorption coefficients; return a array of size (1, 1) for reverberation time;
    - Raise an error for other types of inputs
    Outputs:
    - datain:
    1. "impedance": impedance, shape (6, 1)
    2. "absorpCoefficient": absorption coefficients, shape (6, 1)
    3. "reverberationTime": reverberation time, shape (1, 1)
    """
    dataout = None
    if datatype == "impedance" or datatype == "absorpCoefficient":
        if isinstance(datain, (int, float)):
            dataout = np.full((6, 1), datain)
        elif isinstance(datain, list):
            # If the impedance is a list of single values, convert it to a numpy array
            dataout = np.array(datain)[:, None]
        else:
            raise ValueError("Invalid data shape for {datatype}!")
    elif datatype == "reverberationTime":
        if isinstance(datain, (int, float)):
            dataout = np.full((1, 1), datain)
        else:
            raise ValueError("Invalid data shape for {datatype}!")
    return dataout


def update_n1_n2_n3(params):
    """Update the n1, n2, n3 based on the room size and the sound speed"""
    params["n1"] = int(
        np.floor(
            params["soundSpeed"]
            * params["reverberationTime"]
            / params["roomSize"][0]
            / 2
        )
    )
    params["n2"] = int(
        np.floor(
            params["soundSpeed"]
            * params["reverberationTime"]
            / params["roomSize"][1]
            / 2
        )
    )
    params["n3"] = int(
        np.floor(
            params["soundSpeed"]
            * params["reverberationTime"]
            / params["roomSize"][2]
            / 2
        )
    )
    return params


def compute_rest_params(params):
    """
    Compute the rest of the parameters based on the initialized parameters.
    """
    if params["mode"] == "RIR":  # Add 1/T60 as spacing later !!!
        params["nSamples"] = int(params["sampleRate"] * params["RIRLength"])
        params["freqs"] = np.linspace(
            0,
            params["sampleRate"] / 2,
            round(params["RIRLength"] * params["sampleRate"] / 2) + 1,
        )  # Frequencies will be computed based on the RIR length
        # remove the first DC frequency
        params["freqs"] = params["freqs"][1:]
    elif params["mode"] == "RTF":
        params["freqs"] = np.arange(
            params["startFreq"],
            params["endFreq"] + params["freqStep"],
            params["freqStep"],
        )
    params["waveNumbers"] = (
        2 * np.pi * params["freqs"] / params["soundSpeed"]
    )  # wavenumbers
    # ------------------------------------------------------------
    # For the impedance, we may have a few cases
    # 1. input is just a single value, which means the impedance is uniform in frequency and walls
    # 2. input is a list of single values, which means the impedance is non-uniform in walls but uniform in frequency
    # 3. input is a 2D array of size (6, len(params["freqs"])), which means the impedance is non-uniform in frequency and walls
    # We want to make sure the impedance is a 2D array of size (6, len(params["freqs"]))
    # # : for convex rooms? For other data types?
    # if isinstance(params["impedance"], (int, float)):
    #     params["impedance"] = np.full((6, len(params["freqs"])), params["impedance"])
    # elif isinstance(params["impedance"], list):
    #     # If the impedance is a list of single values, we need to repeat the value for each frequency
    #     params["impedance"] = np.tile(
    #         np.array(params["impedance"])[:, None], (1, len(params["freqs"]))
    #     )
    # elif isinstance(params["impedance"], np.ndarray):
    #     if params["impedance"].shape == (6,):
    #         params["impedance"] = np.tile(
    #             params["impedance"][:, None], (1, len(params["freqs"]))
    #         )
    #     elif params["impedance"].shape == (6, len(params["freqs"])):
    #         pass
    #     else:
    #         raise ValueError("Invalid impedance array shape")
    # #
    # ------------------------------------------------------------
    if params["ifReceiverNormalize"] == 1:
        params["pointSrcStrength"] = (
            1j
            * params["waveNumbers"]
            * params["soundSpeed"]
            * params["airDensity"]
            * params["qFlowStrength"]
        )  # point source strength compensation

    # params["cTs"] = params["soundSpeed"] / params["sampleRate"]
    # try:
    #     # In case of empty roomSize, use "try" to avoid error !!!
    #     L = params["roomSize"] / params["cTs"]
    #     params["n1"], params["n2"], params["n3"] = np.ceil(
    #         params["nSamples"] / (2 * L)
    #     ).astype(int)
    # except:
    #     pass

    return params


def detect_conflicts(params):
    """
    This function detects conflicts between the parameters, you don't have to use it
    But you can place it after you initialize all the parameters before you run the any functional algorithms including:
    - pre_calc_images_src_rec
    - run_DEISM
    - pre_calc_Wigner
    - init_source_directivity
    - init_receiver_directivity
    - And also the other functions used in DEISM-ARG
    """
    # If monopole is used, set spherical harmonic order to 0 for source and receiver
    if params["sourceType"] == "monopole":
        params["sourceOrder"] = 0
    if params["receiverType"] == "monopole":
        params["receiverOrder"] = 0
        params["ifReceiverNormalize"] = 0
    # If not monopole, raise a warning if the spherical harmonic order is also 0
    if params["sourceType"] != "monopole" and params["sourceOrder"] == 0:
        print(
            "[Warning] Spherical harmonic order is set to 0 for source, but source type is not monopole! \n"
        )
    if params["receiverType"] != "monopole" and params["receiverOrder"] == 0:
        print(
            "[Warning] Spherical harmonic order is set to 0 for receiver, but receiver type is not monopole! \n"
        )
    if params["receiverType"] != "monopole" and params["ifReceiverNormalize"] == 0:
        print(
            "[Warning] Receiver normalization is set to 0 for non-monopole receiver, make sure you know what you are doing! \n"
        )
    # If either source or receiver is not monopole, check the distance between source and receiver such that
    # The distance is larger than the radius of the source or receiver
    if params["sourceType"] != "monopole":
        if params["radiusSource"] is None:
            raise ValueError("Source radius is not set for non-monopole source")
        # If distance between source and receiver is smaller than the source radius, raise a warning
        if (
            np.linalg.norm(params["posSource"] - params["posReceiver"])
            < params["radiusSource"]
        ):
            print(
                "[Warning] Distance between source and receiver is smaller than the source radius! \n"
            )
    if params["receiverType"] != "monopole":
        if params["radiusReceiver"] is None:
            raise ValueError("Receiver radius is not set for non-monopole receiver")
        # If distance between source and receiver is smaller than the receiver radius, raise a warning
        if (
            np.linalg.norm(params["posSource"] - params["posReceiver"])
            < params["radiusReceiver"]
        ):
            print(
                "[Warning] Distance between source and receiver is smaller than the receiver radius! \n"
            )
    # If both source and receiver are not monopole, check the distance between source and receiver such that
    # The distance is larger than the radius of the source and receiver
    if params["sourceType"] != "monopole" and params["receiverType"] != "monopole":
        if (
            np.linalg.norm(params["posSource"] - params["posReceiver"])
            < params["radiusSource"] + params["radiusReceiver"]
        ):
            print(
                "[Warning] Distance between source and receiver is smaller than the sum of the source and receiver radius! \n"
            )

    # To be continued...
    # ------------------------------------------------------------
    # Check definitions of wall materials
    if len(params["givenMaterials"]) > 1:
        raise ValueError(
            "The user can only define one of the following three parameters: impedance, absorption coefficient, or reverberation time. The following parameters are defined: "
            + ", ".join(params["givenMaterials"])
        )


def readYaml(filePath):
    """
    This function reads the yaml file and returns a dictionary with the same structure as the yaml file
    inputs:
    - filePath: the address of the yaml file to be read, including the file name with the suffix
    outputs:
    - yaml->dict: a dictionary with the same structure as the yaml file
    """
    # First determine if the file path exists
    # Either in the .test/ directory or in the specified directory
    # Use try to find the file in the tests/ directory
    # if it is found, add tests/ to the file path
    try:
        # If the file is found, assign the file path to filePath
        if fnmatch.filter(os.listdir("examples"), filePath) is not None:
            filePath = "examples/" + filePath
        # If the file is not found, an exception is raised
    except:
        # If the file is not found, assign the file path to filePath
        filePath = filePath
    # Then determine if the file path exists
    if not os.path.exists(filePath):
        # If it does not exist, an exception is thrown
        raise FileExistsError(f"{filePath} doesn't exist!")

    # Then determine if it is a file
    if not os.path.isfile(filePath):
        # raise an exception if it is not a file
        raise FileNotFoundError(f"{filePath} is not a file!")

    # Load yaml normally and return
    with open(filePath, "r") as stream:
        configs = yaml.safe_load(stream)
    # The arrays read directly from yaml are in list format, so they are converted to numpy format
    # Convert the arrays in it to numpy arrays
    for key, value in configs.items():
        if isinstance(value, list):
            configs[key] = np.array(value)

    return configs


def parseCmdArgs(mode="RTF"):
    """
    Add command line parameter modification support, that is, allow some loaded parameters to be modified directly from the command line
    """

    parse = argparse.ArgumentParser(
        description="Please input your self-defined initialized parameters"
    )
    # print(parse.description)
    # Environment parameters
    parse.add_argument(
        "-c", metavar="c0", help="speed of sound(m/s:typical 343)", type=float
    )
    parse.add_argument("-rho", metavar="rho0", help="constant of air", type=float)
    # Room parameters
    parse.add_argument(
        "-room",
        metavar=("Lx", "Ly", "Lz"),
        help='shoebox room size with input format \
        "-room 3.2 4.1 5.6" (m)',
        nargs=3,
        type=float,
    )
    # Reflections
    parse.add_argument("-nro", help="maximum reflection order, integer", type=int)
    # ------------------------------------------------------------
    # The user can only define one of the following three parameters, e.g., -zs or -absp or -t60, if more than one is defined, a warning will be raised
    # Impedance of the walls, can be complex number or a list of complex numbers or a string or a list of strings
    parse.add_argument(
        "-zs",
        metavar=("Z_x1", "Z_x2", "Z_y1", "Z_y2", "Z_z1", "Z_z2"),
        help="acoustic impedance of the the six walls at all frequencies",
        nargs=6,
        type=float,
    )
    parse.add_argument(
        "-absp",
        metavar=(
            "alpha_x1",
            "alpha_x2",
            "alpha_y1",
            "alpha_y2",
            "alpha_z1",
            "alpha_z2",
        ),
        help="absorption coefficients of the the six walls at all frequencies",
        nargs=6,
        type=float,
    )
    parse.add_argument(
        "-t60",
        metavar=("T60"),
        help="reverberation time (positive value) of all frequencies",
        nargs=1,
        type=float,
    )
    # ------------------------------------------------------------
    # Some problems may occur when setting bool parameters !!!
    parse.add_argument(
        "-adrc",
        help="reflection coefficient is angle \
        dependent or not(0:independent, 1:dependent)",
        type=int,
    )
    # Positions of the source and receiver
    parse.add_argument(
        "-xs",
        metavar=("x", "y", "z"),
        help='source position in 3D Cartesian \
        coordinate(m),input format "-xs 2.2 2.1 1.3"',
        nargs=3,
        type=float,
    )
    parse.add_argument(
        "-xr",
        metavar=("x", "y", "z"),
        help='receiver position in 3D Cartesian \
        coordinate(m),input format "-xr 1.2 1.1 2.3"',
        nargs=3,
        type=float,
    )
    # Orientations of the source and receiver
    parse.add_argument(
        "-so",
        metavar=("alpha", "beta", "gamma"),
        help='source orientation in Euler angles \
        Degrees, input format "-so 0 0 0"',
        nargs=3,
        type=float,
    )
    parse.add_argument(
        "-ro",
        metavar=("alpha", "beta", "gamma"),
        help='receiver orientation in Euler angles \
        Degrees, input format "-ro 0 0 0"',
        nargs=3,
        type=float,
    )
    # Frequency parameters
    if mode == "RTF":
        parse.add_argument("-fmin", help="start frequency(Hz)", type=float)
        parse.add_argument("-fstep", help="frequence step size(Hz)", type=float)
        parse.add_argument("-fmax", help="stop frequency(Hz)", type=float)
    elif mode == "RIR":
        parse.add_argument("-fs", help="sampling rate", type=int)
        parse.add_argument("-K", help="over sampling factor", type=float)
        parse.add_argument("-rirlen", help="RIR length (sec)", type=float)
    # -------------------Directivity parameters-------------------
    # Source and receiver directivities types
    parse.add_argument("-srctype", help="source type, string", type=str)
    parse.add_argument("-rectype", help="receiver type, string", type=str)
    # Source directivities parameters
    parse.add_argument(
        "-srcorder",
        help="maximum spherical harmonic \
        directivity order in source",
        type=int,
    )
    parse.add_argument(
        "-srcr0", help="radius of transparent sphere of source(m)", type=float
    )
    # Receiver directivity parameters
    parse.add_argument(
        "-recorder",
        help="maximum spherical harmonic \
        directivity order in receiver",
        type=int,
    )
    parse.add_argument(
        "-recr0", help="radius of transparent sphere of receiver(m)", type=float
    )
    # -------------------Other DEISM parameters-------------------
    # parse.add_argument('-nos',help='number of sample points',type=int)
    # # parse.add_argument('-brc',help='beta reference coefficients',type=float)
    parse.add_argument(
        "-ird",
        help="If remove the direct path or not(0: not remove, 1: remove)",
        type=int,
    )
    parse.add_argument(
        "-method",
        help="Speficy which DEISM mode to use: \
        ORG, LC, MIX",
        type=str,
    )
    parse.add_argument(
        "-meo",
        help="Max. reflection order used for DEISM-ORG: \
        in DEISM-MIX mode",
        type=int,
    )
    parse.add_argument(
        "-npi",
        help="Number of images in parallel computation",
        type=int,
    )
    parse.add_argument(
        "-irn",
        help="If normalize the receiver directivities \
        or not(0: not normalize, 1: normalize)",
        type=int,
    )
    parse.add_argument(
        "-q",
        help="point source flow strength used in \
        obtaining receive directivities",
        type=float,
    )

    # parse.add_argument('-spm',help='sampling scheme',type=str)
    # -------------------trigger running--------------------
    parse.add_argument(
        "--run",
        action="store_true",
        help="Flag to indicate whether to run the DEISM function",
    )
    # Trigger no output messages in the command line
    parse.add_argument(
        "--quiet",
        action="store_true",
        help="Flag to indicate whether to output messages in the command line",
    )
    return parse.parse_args()


def parseCmdArgs_ARG(mode="RTF"):
    """
    Add command line parameter modification support, that is, allow some loaded parameters to be modified directly from the command line
    Todos:
    1. now for the wall materials, only impedance and absorption coefficients are supported by setting the same value for all walls, should support T60 later
    """

    parse = argparse.ArgumentParser(
        description="Please input your self-defined initialized parameters"
    )
    # print(parse.description)
    # Environment parameters
    parse.add_argument(
        "-c", metavar="c0", help="speed of sound(m/s:typical 343)", type=float
    )
    parse.add_argument("-rho", metavar="rho0", help="constant of air", type=float)
    # Room parameters
    # Reflections
    parse.add_argument("-nro", help="maximum reflection order, integer", type=int)
    # Impedance of the walls, can be complex number or a list of complex numbers or a string or a list of strings
    parse.add_argument(
        "-zs",
        metavar=("Z_S"),
        help="acoustic impedance of the all walls at all frequencies",
        nargs=1,
        type=float,
    )
    parse.add_argument(
        "-absp",
        metavar=("alpha_S"),
        help="absorption coefficients of the all walls at all frequencies",
        nargs=1,
        type=float,
    )
    # Positions of the source and receiver
    parse.add_argument(
        "-xs",
        metavar=("x", "y", "z"),
        help='source position in 3D Cartesian \
        coordinate(m),input format "-xs 2.2 2.1 1.3"',
        nargs=3,
        type=float,
    )
    parse.add_argument(
        "-xr",
        metavar=("x", "y", "z"),
        help='receiver position in 3D Cartesian \
        coordinate(m),input format "-xr 1.2 1.1 2.3"',
        nargs=3,
        type=float,
    )
    # Orientations of the source and receiver
    parse.add_argument(
        "-so",
        metavar=("alpha", "beta", "gamma"),
        help='source orientation in Euler angles \
        Degrees, input format "-so 0 0 0"',
        nargs=3,
        type=float,
    )
    parse.add_argument(
        "-ro",
        metavar=("alpha", "beta", "gamma"),
        help='receiver orientation in Euler angles \
        Degrees, input format "-ro 0 0 0"',
        nargs=3,
        type=float,
    )
    # Frequency parameters
    if mode == "RTF":
        parse.add_argument("-fmin", help="start frequency(Hz)", type=float)
        parse.add_argument("-fstep", help="frequence step size(Hz)", type=float)
        parse.add_argument("-fmax", help="stop frequency(Hz)", type=float)
    elif mode == "RIR":
        parse.add_argument("-fs", help="sampling rate", type=int)
        parse.add_argument("-K", help="over sampling factor", type=float)
        parse.add_argument("-rirlen", help="RIR length (sec)", type=float)
    # -------------------Directivity parameters-------------------
    # Source and receiver directivities types
    parse.add_argument("-srctype", help="source type, string", type=str)
    parse.add_argument("-rectype", help="receiver type, string", type=str)
    # Source directivities parameters
    parse.add_argument(
        "-srcorder",
        help="maximum spherical harmonic \
        directivity order in source",
        type=int,
    )
    parse.add_argument(
        "-srcr0", help="radius of transparent sphere of source(m)", type=float
    )
    # Receiver directivity parameters
    parse.add_argument(
        "-recorder",
        help="maximum spherical harmonic \
        directivity order in receiver",
        type=int,
    )
    parse.add_argument(
        "-recr0", help="radius of transparent sphere of receiver(m)", type=float
    )
    # -------------------Other DEISM parameters-------------------
    # parse.add_argument('-nos',help='number of sample points',type=int)
    # # parse.add_argument('-brc',help='beta reference coefficients',type=float)
    parse.add_argument(
        "-ifconvex",
        help="If the room is convex or not(0: not convex, 1: convex)",
        type=int,
    )
    parse.add_argument(
        "-ird",
        help="If remove the direct path or not(0: not remove, 1: remove)",
        type=int,
    )
    parse.add_argument(
        "-method",
        help="Speficy which DEISM mode to use: \
        ORG, LC, MIX",
        type=str,
    )
    parse.add_argument(
        "-meo",
        help="Max. reflection order used for DEISM-ORG: \
        in DEISM-MIX mode",
        type=int,
    )
    parse.add_argument(
        "-npi",
        help="Number of images in parallel computation",
        type=int,
    )
    parse.add_argument(
        "-irn",
        help="If normalize the receiver directivities \
        or not(0: not normalize, 1: normalize)",
        type=int,
    )
    parse.add_argument(
        "-q",
        help="point source flow strength used in \
        obtaining receive directivities",
        type=float,
    )

    # parse.add_argument('-spm',help='sampling scheme',type=str)
    # -------------------trigger running--------------------
    parse.add_argument(
        "--run",
        action="store_true",
        help="Flag to indicate whether to run the DEISM function",
    )
    # Trigger no output messages in the command line
    parse.add_argument(
        "--quiet",
        action="store_true",
        help="Flag to indicate whether to output messages in the command line",
    )
    return parse.parse_args()


def loadSingleParam(configs, args, mode="RTF", roomtype="shoebox"):
    """
    Directly read the configSingleParam.yml file to obtain parameters,
    without first writing the parameters to yml,
    and then reconfiguring the parameters by reading the written yml file.
    Determine whether the command line has entered some parameters.
    If so, override the corresponding parameters in yaml.
    If not, keep the configuration values in yaml
    inputs:
    - args:    argparse library parsed command line object, usually the return value of the parse.parse_args() method
    - configs: a dictionary with the same data in the yaml file
    outputs:
    - params:   the final configuration dictionary, some parameters that need to be calculated
    are not calculated at the current location, because the calculated value is a very long
    """

    if not isinstance(args, argparse.Namespace):
        # Type check is not correct
        raise TypeError(f"args type:{type(args)} is not of type argparse.Namespace.")

    if not isinstance(configs, dict):
        raise TypeError(f"configs is not of type dict")

    # First define params as a dictionary
    params = dict()

    # Then determine whether to use the values of the command line to override
    # Environment parameters, make sure the values are all float
    params["soundSpeed"] = args.c or float(configs["Environment"]["soundSpeed"])
    params["airDensity"] = args.rho or float(configs["Environment"]["airDensity"])
    # ------------------------------------------------------------
    # Room geometry parameters
    # Shoebox room: roomSize
    # Convex room: vertices, wallCenters
    if roomtype == "shoebox":
        params["roomSize"] = np.array(args.room or list(configs["Dimensions"].values()))
    elif roomtype == "convex":
        # TODO: Right now we don't support inputing geometry throught the command line
        # So we use the geometry including vertices and wallCenters defined in the config file
        params["vertices"] = np.array(list(configs["Dimensions"]["vertices"]))
        params["wallCenters"] = np.array(list(configs["Dimensions"]["wallCenters"]))
        params["ifRotateRoom"] = configs["Dimensions"]["ifRotateRoom"]
        params["roomRotation"] = np.array(
            configs["Dimensions"]["roomRotation"].values()
        )
        params["convexRoom"] = args.ifconvex or configs["Dimensions"]["ifConvexRoom"]

    else:
        raise ValueError(
            f"Invalid room type: {roomtype}, must be 'shoebox' or 'convex'"
        )
    # ------------------------------------------------------------
    # Reflections
    # The maximum reflection order is either defined or set to -1, if not defined
    if args.nro is not None or configs["Reflections"]["maxReflectionOrder"] is not None:
        params["maxReflOrder"] = (
            args.nro or configs["Reflections"]["maxReflectionOrder"]
        )
    else:
        params["maxReflOrder"] = -1
    # Impedance, float or list of floats, !!! Should support string or list of strings
    givenMaterials = []
    try:
        params["impedance"] = args.zs or configs["Reflections"]["impedance"]
        givenMaterials.append("impedance")
    except:
        pass
    try:
        params["absorpCoefficient"] = (
            args.absp or configs["Reflections"]["absorpCoefficienticient"]
        )
        givenMaterials.append("absorpCoefficient")
    except:
        pass
    try:
        params["reverberationTime"] = (
            args.t60 or configs["Reflections"]["reverberationTime"]
        )
        givenMaterials.append("reverberationTime")
    except:
        pass
    if len(givenMaterials) > 1:
        raise ValueError(
            "The user can only define one of the following three parameters: -zs or -absp or -t60. The following parameters are defined: "
            + ", ".join(givenMaterials)
        )
    params["givenMaterials"] = givenMaterials
    # Show which material in
    # if len(materials) > 0:
    #     if materials[0] == "impedance":
    #         print("The following material is defined: Acoustic Impedance")
    #     elif materials[0] == "absorpCoefficient":
    #         print("The following material is defined: Absorption Coefficient")
    #     elif materials[0] == "reverberationTime":
    #         print("The following material is defined: Reverberation Time")
    # ------------------------------------------------------------
    # Angle dependent flag, could be 0 or 1 or not defined
    try:
        params["angDepFlag"] = args.adrc or configs["Reflections"]["angleDependentFlag"]
    except:
        pass
    # params["angDepFlag"] = (
    #     args.adrc
    #     if args.adrc is not None
    #     else configs["Reflections"]["angleDependentFlag"]
    # )

    # positions of the source and receiver
    params["posSource"] = np.array(
        args.xs or list(configs["Positions"]["source"].values())
    )
    params["posReceiver"] = np.array(
        args.xr or list(configs["Positions"]["receiver"].values())
    )
    params["orientSource"] = np.array(
        args.so or list(configs["Orientations"]["source"].values())
    )
    params["orientReceiver"] = np.array(
        args.ro or list(configs["Orientations"]["receiver"].values())
    )
    # Frequency parameters
    if mode == "RTF":
        params["startFreq"] = args.fmin or configs["Frequencies"]["startFrequency"]
        params["freqStep"] = args.fstep or configs["Frequencies"]["frequencyStep"]
        params["endFreq"] = args.fmax or configs["Frequencies"]["endFrequency"]
    elif mode == "RIR":
        params["sampleRate"] = args.fs or configs["Signal"]["samplingRate"]
        params["RIRLength"] = args.rirlen or configs["Signal"]["RIRLength"]
        params["overSamplingFactor"] = args.K or configs["Signal"]["overSamplingFactor"]
    params["mode"] = mode

    # Directivity parameters
    params["sourceOrder"] = (
        args.srcorder or configs["MaxSphDirectivityOrder"]["sourceOrder"]
    )
    params["receiverOrder"] = (
        args.recorder or configs["MaxSphDirectivityOrder"]["receiverOrder"]
    )
    params["radiusSource"] = args.srcr0 or configs["Radius"]["source"]
    params["radiusReceiver"] = args.recr0 or configs["Radius"]["receiver"]
    params["sourceType"] = args.srctype or configs["Directivities"]["source"]
    params["receiverType"] = args.rectype or configs["Directivities"]["receiver"]

    params["ifRemoveDirectPath"] = args.ird or configs["DEISM_specs"]["ifRemoveDirect"]
    params["DEISM_method"] = args.method or configs["DEISM_specs"]["Method"]
    try:
        params["mixEarlyOrder"] = args.meo or configs["DEISM_specs"]["mixEarlyOrder"]
    except:
        pass
    params["numParaImages"] = args.npi or configs["DEISM_specs"]["numParaImages"]
    params["ifReceiverNormalize"] = (
        args.irn or configs["DEISM_specs"]["ifReceiverNormalize"]
    )
    params["qFlowStrength"] = args.q or configs["DEISM_specs"]["QFlowStrength"]
    params["silentMode"] = args.quiet or configs["SilentMode"]
    # variables computed according to the above parameters
    # params = compute_rest_params(params)
    return params


def printDict(dict):
    """
    Output the variable of the params dictionary
    """
    if not dict["silentMode"]:
        print("[Parameters]: ", end="\n")
        # Copy the dictionary to avoid modifying the original dictionary
        dict1 = dict.copy()
        # Exclude the following keys
        # Remove the following keys
        excludeKeys = [
            "waveNumbers",
            "pointSrcStrength",
            "n1",
            "n2",
            "n3",
            "nSamples",
            "silentMode",
            "posSources",
            "posReceivers",
            "orientSources",
            "orientReceivers",
            "cTs",
        ]
        # If the source is monopole, remove radiusSource
        if dict["sourceType"] == "monopole":
            excludeKeys.append("radiusSource")
        # If the receiver is monopole, remove radiusReceiver
        if dict["receiverType"] == "monopole":
            excludeKeys.append("radiusReceiver")
        for key in excludeKeys:
            if key in dict1.keys():
                dict1.pop(key)
        # First traverse to find the longest key for the purpose of providing formatted output
        maxLen = 0
        # For all parameter names
        for key in dict1.keys():
            # If the length of the key is greater than maxLen, then maxLen is equal to the length of the key
            maxLen = maxLen if len(key) < maxLen else len(key)
        # For all parameter name-value pairs
        for key, value in dict1.items():
            # If the value is a list and the length of the value is greater than 6
            # Or the value is a 2D numpy array
            if (isinstance(value, list) and len(value) > 6) or isinstance(
                value, np.ndarray
            ):
                # If print the 2D array value of the key "vertices",
                if key == "vertices":
                    # If "if_rotate_room" is 1, add "rotated" before the key "vertices"
                    if dict["ifRotateRoom"] == 1:
                        key = "rotated vertices"
                    # For each row in the 2D array
                    row_id = 0
                    for row in value:
                        # Output the key with incrementing ids and the row
                        row_id += 1
                        # Round the row to 2 decimal places
                        row = np.round(row, 2)
                        print(f"{key:>{maxLen}} {row_id} : {row}", end="\n")
                    # Skip the following output
                    continue
                # If the key is acousImpend with 2D arrays, output the key and the 2D array separately
                # Each entry's name is Impedance wall 1, Impedance wall 2, etc.
                if key == "impedance":
                    key = "Impedance wall "
                    # For each row in the 2D array
                    row_id = 0
                    for row in value:
                        # Output the key with incrementing ids and the row
                        row_id += 1
                        # Round the row to 2 decimal places
                        row = np.round(row, 2)
                        print(
                            f"{key:>{maxLen}} {row_id} : {row[:2]} ... {row[-2:]}",
                            end="\n",
                        )
                    # Skip the following output
                    continue
                # The value is truncated to the first two elements and last two elements
                if isinstance(value, list):
                    valueStr = f"{value[:2]} ... {value[-2:]}"
                elif isinstance(value, np.ndarray) and value.ndim == 1:
                    if len(value) < 6:
                        valueStr = f"{value}"
                    else:
                        valueStr = f"{value[:2]} ... {value[-2:]}"

            else:
                # Otherwise, the value is converted to a string
                valueStr = str(value)
            # Output the key and value
            print(f"{key:>{maxLen+2}} : {valueStr}", end="\n")
        # Print some additional parameters individually
        # If vertices is in the dictionary, output the vertices (2D array) separately on the right side
        # if "vertices" in dict.keys():
        #     print("vertices : ", dict["vertices"])

        # Other messages, notes for the users:
        print("[Notes]: ", end="\n")
        print(
            "If the receiver directivity is obtained by placing a point source with flow strength QFlowStrength ",
            end="",
        )  # Rephrase the sentences and delete the spaces between lines
        print(
            "at the receiver position and measuring the sound pressure at a sphere around the receiver, ",
            end="",
        )
        print("you need to specify QFlowStrength and set ifReceiverNormalize to 1. \n")
        print(
            "If the receiver directivity is obtained by other methods, you can set ifReceiverNormalize to 0. ",
            end="\n\n",
        )
        # print("[All] ", end="\n")


def cmdArgsToDict(mode="RTF", roomtype="shoebox"):
    """
    Takes the command line arguments and the parameters from the configSingleParam.yml file
    outputs:
    - params: the final configuration dictionary
    - cmdArgs: the command line arguments
    """
    # Decide with default yml name to load
    if roomtype == "shoebox":
        if mode == "RTF":
            yml_name = "configSingleParam_RTF.yml"
        elif mode == "RIR":
            yml_name = "configSingleParam_RIR.yml"
        else:
            raise ValueError(f"Invalid mode: {mode}")
    elif roomtype == "convex":
        if mode == "RTF":
            yml_name = "configSingleParam_ARG_RTF.yml"
        elif mode == "RIR":
            yml_name = "configSingleParam_ARG_RIR.yml"
        else:
            raise ValueError(f"Invalid mode: {mode}")
    else:
        raise ValueError(f"Invalid room: {roomtype}, must be 'shoebox' or 'convex'")
    # First load the parameters in the configSingleParam.yml file directly as params,
    configsInYaml = readYaml(yml_name)
    # parse the command line arguments
    # TODO: merge the two functions parseCmdArgs and parseCmdArgs_ARG later
    if roomtype == "shoebox":
        cmdArgs = parseCmdArgs(mode)
    elif roomtype == "convex":
        cmdArgs = parseCmdArgs_ARG(mode)
    else:
        raise ValueError(f"Invalid room: {roomtype}, must be 'shoebox' or 'convex'")
    # replace the corresponding variables in configsInYaml with the values entered from the command line
    params = loadSingleParam(configsInYaml, cmdArgs, mode, roomtype)
    # Compute the rest of the parameters
    # params = compute_rest_params(params)

    return params, cmdArgs


# def cmdArgsToDict_ARG(mode="RTF"):
#     """
#     Takes the command line arguments and the parameters from the configSingleParam.yml file
#     outputs:
#     - params: the final configuration dictionary
#     - cmdArgs: the command line arguments
#     """
#     # Decide with default yml name to load
#     if mode == "RTF":
#         yml_name = "configSingleParam_ARG_RTF.yml"
#     elif mode == "RIR":
#         yml_name = "configSingleParam_ARG_RIR.yml"
#     else:
#         raise ValueError(f"Invalid mode: {mode}")
#     # First load the parameters in the configSingleParam.yml file directly as params,
#     configsInYaml = readYaml(yml_name)
#     # parse the command line arguments
#     cmdArgs = parseCmdArgs_ARG(mode)
#     # replace the corresponding variables in configsInYaml with the values entered from the command line
#     params = loadSingleParam(configsInYaml, cmdArgs, mode)
#     # Compute the rest of the parameters
#     params = compute_rest_params(params)

#     return params, cmdArgs


def load_directive_pressure(silentMode, src_or_rec, name):
    """
    Functions for loading sampled directional pressure field on a sphere with radius r0 simulated from COMSOL or simulation
    input:
        src_or_rec: string, "source" or "receiver"
        name: string, the name of the profile
    output:
        freqs_all: 1D array, the frequencies of the simulated pressure field
        pressure: 2D array, number of frequencies x number of sample points, the simulated pressure field
        Dir_all: 2D array, number of directions x [azimuth, elevation], the sample points of directivities (pressure field) on the sphere
        r0: 1D array, the radius of the transparent sphere
    """
    path = "data/sampled_directivity"
    # Use try to find the file in the tests/ directory
    # if it is found, add tests/ to the file path
    try:
        # If the file is found, assign the file path to filePath
        if os.path.exists("examples/data"):
            path = "./examples/" + path
        elif os.path.exists("data"):
            # do nothing
            pass
        # If the path is not found, an exception is raised
    except FileNotFoundError(f"{path} doesn't exist!"):
        # stop the program and print the error message
        pass
    data_location = "{}/{}/{}.mat".format(path, src_or_rec, name)
    try:
        with open(data_location, "rb") as file:
            FEM_data = sio.loadmat(file)
    except FileNotFoundError:
        raise FileNotFoundError(f"File {data_location} not found.")
    except Exception as e:
        raise RuntimeError(f"An error occurred while loading the file: {e}")

    freqs_all = FEM_data["freqs_mesh"].flatten()
    pressure = FEM_data["Psh"]
    Dir_all = FEM_data["Dir_all"]
    r0 = FEM_data["r0"]
    if not silentMode:
        print("Load directivity data from {}, ".format(path), end="")
        print(
            f"sampled on {len(Dir_all)} directions and {len(freqs_all)} frequencies. ",
            end="",
        )

    return freqs_all, pressure, Dir_all, r0


def load_directpath_pressure(silentMode, name):
    """
    Function that loads the direct path pressure field simulated from COMSOL
    """
    path = "data/sampled_directivity"
    # Use try to find the file in the tests/ directory
    # if it is found, add tests/ to the file path
    try:
        # If the file is found, assign the file path to filePath
        if os.path.exists("examples/data"):
            path = "./examples/" + path
        elif os.path.exists("data"):
            # do nothing
            pass
        # If the path is not found, an exception is raised
    except FileNotFoundError(f"{path} doesn't exist!"):
        # stop the program and print the error message
        pass
    data_location = "{}/source/{}.mat".format(path, name)
    try:
        with open(data_location, "rb") as file:
            FEM_data = sio.loadmat(file)
    except FileNotFoundError:
        raise FileNotFoundError(f"File {data_location} not found.")
    except Exception as e:
        raise RuntimeError(f"An error occurred while loading the file: {e}")

    freqs_all = FEM_data["freqs_mesh"].flatten()
    pressure = FEM_data["Psh"]
    mic_pos = FEM_data["mic_position"]
    if not silentMode:
        print(f"[Data] Load direct path data from {path}. ", end="")
        print(f"Microphone positions: {mic_pos}. ", end="\n")
    return freqs_all, pressure, mic_pos


def load_RTF_data(silentMode, name):
    """
    This functions loads the room transfer functions (RTFs) simulated from COMSOL
    """
    path = "data/RTF_COMSOL"
    # Use try to find the file in the tests/ directory
    try:
        # If the file is found, assign the file path to filePath
        if os.path.exists("examples/data"):
            path = "./examples/" + path
        elif os.path.exists("data"):
            # do nothing
            pass
        # If the path is not found, an exception is raised
    except FileNotFoundError(f"{path} doesn't exist!"):
        # stop the program and print the error message
        pass

    data_location = "{}/{}.mat".format(path, name)
    try:
        with open(data_location, "rb") as file:
            P_COMSOL = sio.loadmat(file)
    except FileNotFoundError:
        raise FileNotFoundError(f"File {data_location} not found.")
    except Exception as e:
        raise RuntimeError(f"An error occurred while loading the file: {e}")

    # Flatten the 2D array to 1D array
    pressure = P_COMSOL["Psh"]
    try:
        freqs_all = P_COMSOL["freqs_mesh"].flatten()
        mic_pos = P_COMSOL["mic_pos"]
    # If not freqs_mesh and mic_pos entries in the dictionary, return two zeros
    except KeyError:
        freqs_all = np.zeros(1)
        mic_pos = np.zeros(1)
    if not silentMode:
        print(f"[Data] Load RTF data from {path}, Done! ", end="\n")

    return freqs_all, pressure, mic_pos



"""
Recreating the results of Figure 5 and Figure 6 from the following conference paper:
Z. Xu, E. A. P. Habets and A. G. Prinn,
"Simulating Sound Fields in Rooms with Arbitrary Geometries Using the Diffraction-Enhanced Image Source Method,"
2024 18th International Workshop on Acoustic Signal Enhancement (IWAENC), Aalborg, Denmark, 2024, pp. 284-288,
doi: 10.1109/IWAENC61483.2024.10693991.
The following scnario is considered in Figure 5:
1. A room with tilted ceiling compared to a shoebox room
2.
"""

import time
import os
import psutil
import numpy as np
import ray

from deism.core_deism import *
from deism.core_deism_arg import *
from deism.data_loader import *
from deism.utilities import plot_RTFs


def init_parameters_convex_fig5(params):
    # reflection order
    params["maxReflOrder"] = 15
    # A simple room with 8 vertices
    params["vertices"] = np.array(
        [
            [0, 0, 0],  # Origin, but it does not have to be the origin
            [0, 0, 3.5],  # [x, y, z] coordinates of the room vertices
            [0, 3, 2.5],
            [0, 3, 0],
            [4, 0, 0],
            [4, 0, 3.5],
            [4, 3, 2.5],
            [4, 3, 0],
        ]
    )
    params["if_rotate_room"] = 0
    # Source and receiver positions
    params["posSource"] = np.array([1.1, 1.1, 1.3])
    params["posReceiver"] = np.array([2.9, 1.9, 1.3])
    # Orientations of the sources and receivers
    params["orientSources"] = np.array([0, 0, 0])
    params["orientReceiver"] = np.array([180, 0, 0])
    # Radius of the spheres in meters
    params["radiusSource"] = 0.2
    params["radiusReceiver"] = 0.25
    # Source and receiver directivity profiles
    params["sourceType"] = "Speaker_small_sph_cyldriver_source"
    params["receiverType"] = "Speaker_small_sph_cyldriver_receiver"
    # Compute the rest of the parameters
    params = compute_rest_params(params)
    return params


def init_parameters_convex_fig6(params):
    # reflection order
    params["maxReflOrder"] = 15
    # A simple room with 8 vertices
    params["vertices"] = np.array(
        [
            [0, 0, 0],  # Origin
            [0, 0, 3.25],
            [0, 3, 2.75],
            [0, 3, 0],
            [4, 0, 0],
            [4, 0, 3.25],
            [4, 3, 2.75],
            [4, 3, 0],
        ]
    )
    params["if_rotate_room"] = 0
    # Source and receiver positions
    params["posSource"] = np.array([1.1, 1.1, 1.3])
    params["posReceiver"] = np.array([2.9, 1.9, 1.3])
    # Orientations of the sources and receivers
    params["orientSources"] = np.array([0, 0, 0])
    params["orientReceiver"] = np.array([180, 0, 0])
    # Radius of the spheres in meters
    params["radiusSource"] = 0.2
    params["radiusReceiver"] = 0.25
    # Source and receiver directivity profiles
    params["sourceType"] = "Speaker_small_sph_cyldriver_source"
    params["receiverType"] = "Speaker_small_sph_cyldriver_receiver"
    return params


def plot_DEISM_ARG_FEM(P_DEISM_ARG, P_FEM, freqs, save_path, fig_name):
    plt.rcParams["text.usetex"] = True
    # Get the SPL of the DEISM-ARG and FEM results
    SPL_DEISM_ARG = get_SPL(P_DEISM_ARG)
    SPL_FEM = get_SPL(P_FEM)
    # Calculate RMS-LSD
    RMS_LSD = np.sqrt(
        np.sum(np.abs(10 * np.log10((np.abs(P_DEISM_ARG) / np.abs(P_FEM)) ** 2)) ** 2)
        / len(P_DEISM_ARG)
    )
    fig = plt.figure(figsize=(18, 8))
    ax = fig.add_subplot(1, 1, 1)
    ax.plot(freqs, SPL_FEM, label="FEM", color="black", linestyle="-", linewidth=3)
    ax.plot(
        freqs, SPL_DEISM_ARG, label="DEISM-ARG", color="red", linestyle="-", linewidth=3
    )
    ax.set_xlim([freqs[0], freqs[-1]])
    # ax.set_ylim([20, 100])
    ax.set_xlabel(r"\bf{Frequency (Hz)}")
    ax.set_ylabel(r"\bf{SPL (dB)}")
    ax.xaxis.label.set_size(40)
    ax.yaxis.label.set_size(40)
    ax.xaxis.set_tick_params(labelsize=50)
    ax.yaxis.set_tick_params(labelsize=50)
    # Empty plot to add the RMS LSD as a extra line of text to the legend, keep 3 digits after the decimal point
    # Don't show the line
    ax.plot([], [], label="LSD: {:.2f} dB".format(RMS_LSD), color="white")
    # ax.set_xscale("log")
    ax.legend(fontsize=35, loc="best", bbox_to_anchor=(0.6, 0.4))
    plt.grid(axis="both", which="both", linestyle=":")
    fig.tight_layout()
    fig_name = "{}/IWAENC_{}_SPL.png".format(save_path, fig_name)
    plt.savefig(fig_name, dpi=300)


def main():
    # Choose plot figure 5 or 6
    fig = "fig6"  # "fig5" or "fig6"
    # Load the parameters from the configSingleParam_ARG.yaml file amd command line
    params, cmdArgs = cmdArgsToDict_ARG("configSingleParam_arg.yml")
    # Initialize some additional parameters for DEISM-ARG
    # For example, room vertices, room rotation, etc.
    if fig == "fig5":
        params = init_parameters_convex_fig5(params)
    elif fig == "fig6":
        params = init_parameters_convex_fig6(
            params
        )  # init_parameters_fig5 or init_parameters_fig6
    # detect conflicts
    detect_conflicts(params)
    # print the parameters or not
    if cmdArgs.quiet:
        params["silentMode"] = 1
    printDict(params)
    # -------------------------------------------------------
    # Shared parameters for all modes, ORG, LC, MIX
    # -------------------------------------------------------
    params = init_receiver_directivities_ARG(params, params["if_rotate_room"])
    # Using DESIM-ARG with c++ libroom_deism to find the visible images and reflection matrices, attenuations
    # initialize Room_deism
    room_pra_deism = Room_deism_cpp(params)
    # Or you can use the Python version, which is much slower in generating the images
    # room_pra_deism = Room_deism_python(params)
    # Plot the room and the images
    # room_pra_deism.plot_room()
    # get the reflection paths
    params = get_ref_paths_ARG(params, room_pra_deism)
    # Calculating the reflected source directivity coefficients
    params = init_source_directivities_ARG(
        params, params["if_rotate_room"], params["reflection_matrix"]
    )
    # Use the DEISM-ARG version
    Wigner = pre_calc_Wigner(params)
    # -------------------------------------------------------
    # Run DEISM-ARG parallel processing using Ray
    # -------------------------------------------------------
    # Initialize Ray
    num_cpus = psutil.cpu_count(logical=False)
    if not ray.is_initialized():
        ray.init(num_cpus=num_cpus)
        print("\n")
    # Run DEISM-ARG for specified mode
    P_DEISM_ARG = ray_run_DEISM_ARG_ORG(params, params["images"], Wigner)
    # Shutdown Ray
    ray.shutdown()
    # -------------------------------------------------------
    # Load FEM data
    # -------------------------------------------------------
    if fig == "fig5":
        freqs_FEM, P_FEM, mic_pos = load_RTF_data(
            params["silentMode"], "Room_iwaenc_fig5"
        )
    elif fig == "fig6":
        freqs_FEM, P_FEM, mic_pos = load_RTF_data(
            params["silentMode"], "Room_iwaenc_fig6"
        )  # "Room_iwaenc_fig5" or "Room_iwaenc_fig6"
    # -------------------------------------------------------
    # Plot the RTFs
    # -------------------------------------------------------
    save_path = "./outputs/figures"
    # check if the save path exists
    if not os.path.exists(save_path):
        os.makedirs(save_path)
    plot_DEISM_ARG_FEM(P_DEISM_ARG, P_FEM.flatten(), params["freqs"], save_path, fig)


# -------------------------------------------------------
if __name__ == "__main__":
    main()



"""
Comparing the image generation process using:
1. pyroomacoustics
2. deism-arg (c++)
We compare the two methods by comparing
1. The computation time
2. The number of images generated, if they are the same
3. The image positions, if they are the same, or within a certain tolerance
!!!! You may see difference between the two methods, this is due to the numerical precision of the two methods
Difference start to appear at higher reflection orders, e.g, 9 and above
"""

import time
import os
import sys
import psutil
import numpy as np
import ray
import pyroomacoustics as pra
import scipy.special as scy
import matplotlib.pyplot as plt

# from deism.core_deism import

from deism.core_deism_arg import (
    Room_deism_cpp,
    rotate_room_src_rec,
    compare_array_lists_by_distance,
)

from deism.data_loader import (
    cmdArgsToDict_ARG,
    printDict,
    detect_conflicts,
)


def init_parameters_convex(params):
    """
    Initialize some important additional parameters for DEISM-ARG
    In addition to the ones defined in configSingleParam_ARG.yaml
    """
    # Room vertices
    # -------------------------------------------------------
    # Note that we the chosen room vertices and room rotation,
    # We can compare the room created using pyroomacoustics,
    # which is created by extrude the room vertices in the z-direction
    # -------------------------------------------------------
    # Define the room geometry for the DEISM-ARG
    # -------------------------------------------------------
    vertices = np.array(
        [
            [0, 0, 0],  # Origin
            [0, 0, 3.5],
            [0, 3, 2.5],
            [0, 3, 0],
            [4, 0, 0],
            [4, 0, 3.5],
            [4, 3, 2.5],
            [4, 3, 0],
        ]
    )
    # If you need to find the wall centers, use the following code
    # This may be useful for some applications, e.g., bonding the impedance to the walls using the room center
    # wall_centers = find_wall_centers(vertices)
    wallCenters = np.array(
        [
            [0, 1.5, 1.5],
            [2, 3, 1.25],
            [4, 1.5, 1.5],
            [2, 0, 1.75],
            [2, 1.5, 0],
            [2, 1.5, 3],
        ]
    )
    # You need to adlign the acoustic impedance to the wall centers defined above
    # The format of the impedance is [number of walls, len(freqs)], 2D numpy array
    params["acousImpend"] = np.arange(1, 6 * len(params["freqs"]) + 1).reshape(
        6, len(params["freqs"])
    )
    # --- Room rotation, if rotate the room w.r.t the origin ---
    if_rotate_room = 1
    # --- Room rotation angles using Z-X-Z Euler angles ---
    # [alpha, beta, gamma], 3D Euler angles, The rotation matrix calculation used in COMSOL, see:
    # https://doc.comsol.com/5.5/doc/com.comsol.help.comsol/comsol_ref_definitions.12.092.html
    # The chosen rotation angles are used to illustrate a scenario where we can compare with the room created using pyroomacoustics
    # See file DEISM_ARG_comparisons.py
    room_rotation = np.array([90, 90, 90])  # [alpha, beta, gamma] in degrees
    # --- Add the above parameters to the params dictionary ---
    params["vertices"] = vertices
    params["wallCenters"] = wallCenters
    params["if_rotate_room"] = if_rotate_room
    params["room_rotation"] = room_rotation
    # Apply room rotation to the room vertices and source/receiver positions
    if if_rotate_room:
        params = rotate_room_src_rec(params)
    # -------------------------------------------------------
    # Define the room geometry for the pyroomacoustics
    # -------------------------------------------------------
    # Create the same shaped room using 2D polygon
    pol_pra = np.array(
        [[2.5, -3], [3.5, 0], [0, 0], [0, -3]]
    ).T  # * 1.5  # room shape 1, rotated
    # The height of the room
    height_pra = 4
    # Save to the params dictionary
    params["pol_pra"] = pol_pra
    params["height_pra"] = height_pra
    return params


def main():
    params, cmdArgs = cmdArgsToDict_ARG("configSingleParam_arg.yml")
    params = init_parameters_convex(params)
    # detect conflicts
    detect_conflicts(params)
    # print the parameters or not
    if cmdArgs.quiet:
        params["silentMode"] = 1
    # print the parameters
    printDict(params)
    # -------------------------------------------------------
    # If you want to modify the parameters, you can also do it here
    # by changing the values of the parameters in the params dictionary
    # -------------------------------------------------------
    # If run DEISM function, run if --run flag is set in the cmd
    # If cmdArgs are all None values, run following codes directily
    if cmdArgs.run or all(
        value in [None, False] for value in vars(cmdArgs).values()
    ):  # no input in cmd will also run using default parameters
        # -------------------------------------------------------
        # ----- Run image generation using deism-arg ------------
        # -------------------------------------------------------
        begin = time.time()
        # initialize Room_deism
        room_deism_arg = Room_deism_cpp(params)
        # time used
        elapsed_deism_arg = time.time() - begin
        # print(
        #     "Image generation of DEISM-ARG took {} min".format(elapsed_deism_arg / 60)
        # )
        # -------------------------------------------------------
        # ----- Run image generation using pyroomacoustics ------
        # -------------------------------------------------------
        if not cmdArgs.quiet:
            print("[Calculating] pyroomacoustics image generation, ", end="")
        begin = time.time()
        # r_absor = 0.1
        # mat = pra.Material(0.15, 0.1)
        room_pra = pra.Room.from_corners(
            params["pol_pra"],
            fs=16000,
            # absorption=r_absor,
            # materials=mat,
            max_order=params["maxReflOrder"],
            ray_tracing=False,
            air_absorption=False,
        )

        # Create the 3D room by extruding the 2D polygon
        room_pra.extrude(params["height_pra"])
        room_pra.add_source(params["posSource"])
        room_pra.add_microphone(params["posReceiver"])
        room_pra.image_source_model()
        elapsed_pra = time.time() - begin
        if not cmdArgs.quiet:
            minutes, seconds = divmod(elapsed_pra, 60)
            print(f"Done [{minutes} minutes, {seconds:.3f} seconds]", end="\n\n")
            # show the comparison of the two methods' speed
            print(
                "[Info] DEISM-ARG (c++) is {} times faster than pyroomacoustics. ".format(
                    elapsed_pra / elapsed_deism_arg
                ),
                end="\n\n",
            )
        # -------------------------------------------------------
        # Now visualize the room created by pyroomacoustics and DEISM-ARG
        # -------------------------------------------------------
        # # Plot the room created by pyroomacoustics
        # fig, ax = room_pra.plot()
        # ax.set_xlim([0, 12])
        # ax.set_ylim([0, 12])
        # ax.set_zlim([0, 12])
        # plt.show()
        # # Plot the room created by DEISM-ARG
        # room_deism_arg.plot_room()
        # -------------------------------------------------------
        # Now compare the images generated by the two methods
        # -------------------------------------------------------
        # Choose the precision for the comparison
        round_decimal = 3
        # Get the images generated by DEISM-ARG
        images_deism_arg = room_deism_arg.sources.T.round(decimals=round_decimal)
        # Get the images generated by pyroomacoustics
        images_pra = room_pra.sources[0].images.T.round(decimals=round_decimal)
        # -------------------------------------------------------
        # First compare the number of images generated
        # -------------------------------------------------------
        if not cmdArgs.quiet:
            print("[Checking] Image number: ", end="\n")
            print(f"Images generated by pyroomacoustics is {len(images_pra)}, ", end="")
            print(
                f"Images generated by DEISM-ARG (C++) is {len(images_deism_arg)}. ",
                end="\n\n",
            )
            # -------------------------------------------------------
            # Then compare the image positions
            # -------------------------------------------------------
            print(
                "[Checking] If the images are the same between DEISM-ARG (c++) and pyroomacoustics:",
                end="\n",
            )
        # In the function compare_array_lists_by_distance, we show
        # 1. If the images in each method are distinct from each other
        # 2. If the images between the two methods are the same, in the end, unmatch images are returned
        # in "unmatched in xxx" where xxx is the method that has the unmatched images
        # Possible reasons for the mismatches:
        # numerical precision, the two methods may have different numerical precisions,
        # which may lead to slightly different results at higher reflection orders
        mismatches = compare_array_lists_by_distance(
            list(images_pra),
            list(images_deism_arg),
            ["pyroomacoustics", "DEISM-ARG (C++)"],
            tolerance=1e-5,
        )


# -------------------------------------------------------
if __name__ == "__main__":
    main()



"""
Helper functions for room geometry checking and updating.
Contributor:
Anjana
Zeyu Xu
"""

import gmsh
import numpy as np
from scipy.spatial import ConvexHull
import re
import os
import json

# -------------------------------
# Helper functions
# -------------------------------


def load_geometry(filename):
    gmsh.initialize()
    gmsh.open(filename)
    gmsh.model.mesh.generate(3)
    return gmsh


def get_points_and_tets(gmsh_obj):
    node_tags, coords, _ = gmsh_obj.model.mesh.getNodes()
    points = np.array(coords).reshape(-1, 3)

    elem_types, elem_tags, elem_node_tags = gmsh_obj.model.mesh.getElements(dim=3)
    if len(elem_node_tags) == 0:
        raise ValueError("No 3D elements (tetrahedra) found in mesh.")

    tets = np.array(elem_node_tags[0]).reshape(-1, 4) - 1  # zero-based
    return points, tets


def mesh_volume(points, tets):
    vol = 0.0
    for t in tets:
        p0, p1, p2, p3 = points[t]
        vol += abs(np.dot(np.cross(p1 - p0, p2 - p0), p3 - p0)) / 6.0
    return vol


def is_convex(points, tets, tol=1e-2):
    hull = ConvexHull(points)
    hull_vol = hull.volume
    mesh_vol = mesh_volume(points, tets)

    # If hull is much larger → non-convex (like L-shape)
    if (hull_vol - mesh_vol) / hull_vol > tol:
        return False
    return True


def get_corner_points(gmsh_obj):
    entities = gmsh_obj.model.getEntities(dim=0)
    points = []
    for dim, tag in entities:
        coords = gmsh_obj.model.getValue(dim, tag, [])
        points.append(coords)
    return np.array(points).reshape(-1, 3)


def is_shoebox_corners(corners, tol=1e-6):
    x_vals = np.unique(np.round(corners[:, 0] / tol) * tol)
    y_vals = np.unique(np.round(corners[:, 1] / tol) * tol)
    z_vals = np.unique(np.round(corners[:, 2] / tol) * tol)
    return len(x_vals) == 2 and len(y_vals) == 2 and len(z_vals) == 2


# -------------------------------
# Main
# -------------------------------
def get_room_geometry(geo_file):

    gmsh_obj = load_geometry(geo_file)
    points, tets = get_points_and_tets(gmsh_obj)
    volume = mesh_volume(points, tets)

    convex = is_convex(points, tets)
    corners = get_corner_points(gmsh_obj)
    shoebox = is_shoebox_corners(corners)

    if convex:
        room = "convex"
    else:
        room = "shoebox"
    # print("Mesh-based Volume:", volume)
    # print("Convex:", convex)
    # print("Shoebox:", shoebox)

    gmsh_obj.finalize()
    return volume, room


def update_surface_areas(json_file_path, geo_file_path):
    """
    Calculate surface areas from the Gmsh .geo file and write them
    into the JSON under "geometry[0]['room_areas']".
    """
    # Check that files exist
    if not os.path.exists(json_file_path):
        raise FileNotFoundError(f"JSON file not found: {json_file_path}")
    if not os.path.exists(geo_file_path):
        raise FileNotFoundError(f"GEO file not found: {geo_file_path}")

    # Initialize Gmsh
    gmsh.initialize()
    gmsh.open(geo_file_path)
    gmsh.model.mesh.generate(2)  # 2D surface mesh

    # Get all nodes
    node_tags_all, coords_all, _ = gmsh.model.mesh.getNodes()
    coords = coords_all.reshape((len(node_tags_all), 3))

    # Get all surface groups
    surface_group_tags = gmsh.model.getPhysicalGroups(dim=2)
    areas = {}

    for dim, tag in surface_group_tags:
        group_name = gmsh.model.getPhysicalName(dim, tag)
        if not group_name:
            group_name = f"Surface_{tag}"

        elem_types, elem_tags, node_tags = gmsh.model.mesh.getElements(dim, tag)
        if len(node_tags) == 0:
            areas[group_name] = 0.0
            continue

        faces = np.array(node_tags[0]).reshape(-1, 3) - 1  # 0-based
        v0 = coords[faces[:, 0], :]
        v1 = coords[faces[:, 1], :]
        v2 = coords[faces[:, 2], :]
        tri_areas = 0.5 * np.linalg.norm(np.cross(v1 - v0, v2 - v0), axis=1)
        areas[group_name] = float(np.sum(tri_areas))

    gmsh.finalize()

    # Update JSON
    with open(json_file_path, "r") as f:
        data = json.load(f)

    if "geometry" not in data:
        data["geometry"] = [{}]
    elif len(data["geometry"]) == 0:
        data["geometry"].append({})

    data["geometry"][0]["room_areas"] = areas

    with open(json_file_path, "w") as f:
        json.dump(data, f, indent=4)

    print("✅ Surface areas updated in JSON:")
    for name, area in areas.items():
        print(f"  {name}: {area:.2f} m²")

    # return areas


def update_wall_centers(json_path, geo_path):
    with open(geo_path, "r") as f:
        lines_geo = f.readlines()

    points = {}
    point_pattern = re.compile(
        r"Point\((\d+)\) = \{ ([\d\.\-e]+), ([\d\.\-e]+), ([\d\.\-e]+), [\d\.\-e]+ \};"
    )
    for line in lines_geo:
        match = point_pattern.match(line.strip())
        if match:
            pid = int(match.group(1))
            x, y, z = (
                float(match.group(2)),
                float(match.group(3)),
                float(match.group(4)),
            )
            points[pid] = [x, y, z]

    lines_dict = {}
    line_pattern = re.compile(r"Line\((\d+)\) = \{ (\d+), (\d+) \};")
    for line in lines_geo:
        match = line_pattern.match(line.strip())
        if match:
            lid = int(match.group(1))
            lines_dict[lid] = [int(match.group(2)), int(match.group(3))]

    line_loops = {}
    loop_pattern = re.compile(r"Line Loop\((\d+)\) = \{([^\}]+)\};")
    for line in lines_geo:
        match = loop_pattern.match(line.strip())
        if match:
            lid = int(match.group(1))
            line_ids = [int(x.strip()) for x in match.group(2).split(",")]
            line_loops[lid] = line_ids

    physical_surfaces = {}
    phys_pattern = re.compile(r'Physical Surface\("([^"]+)"\) = \{([^\}]+)\};')
    for line in lines_geo:
        match = phys_pattern.match(line.strip())
        if match:
            name = match.group(1)
            plane_ids = [int(x.strip()) for x in match.group(2).split(",")]
            if len(plane_ids) == 1:
                physical_surfaces[name] = plane_ids[0]

    def compute_center(line_ids):
        unique_points = set()
        for l in line_ids:
            l_id = abs(l)
            unique_points.update(lines_dict[l_id])
        x_sum = y_sum = z_sum = 0.0
        for p in unique_points:
            x, y, z = points[p]
            x_sum += x
            y_sum += y
            z_sum += z
        n = len(unique_points)
        return [x_sum / n, y_sum / n, z_sum / n]

    centers = {}
    for name, plane_id in physical_surfaces.items():
        line_loop = line_loops[plane_id]
        centers[name] = compute_center(line_loop)

    with open(json_path, "r") as f:
        data = json.load(f)

    if "geometry" not in data:
        data["geometry"] = [{}]
    # Add room vertices to the JSON
    data["geometry"][0]["vertices"] = [points[i + 1] for i in range(len(points))]
    # Add wall centers to the JSON
    data["geometry"][0]["wall_centers"] = {
        name: f"{c[0]:.4f}, {c[1]:.4f}, {c[2]:.4f}" for name, c in centers.items()
    }

    with open(json_path, "w") as f:
        json.dump(data, f, indent=4)

    print("✅ Wall centers updated successfully!")



import yaml

import argparse
import os
import time
import numpy as np
import matplotlib.pyplot as plt
import ray
from deism.core_deism import *
from deism.data_loader import *
from deism.room_check import (
    get_room_geometry,
    update_surface_areas,
    update_wall_centers,
)

# Import modules
import itertools
import json
import os
import pickle
import time
import types

from math import ceil
from math import log

# Silvin: debugging
import logging
import traceback


logger = logging.getLogger(__name__)


def parse_value(val):
    """Handle strings of comma-separated floats OR single float values."""
    if isinstance(val, str):
        return np.array([float(x.strip()) for x in val.split(",") if x.strip()])
    elif isinstance(val, (int, float)):
        return np.array([val])
    elif isinstance(val, (list, tuple)):
        return np.array(val, dtype=float)
    else:
        raise ValueError(f"Unsupported type for parse_value: {type(val)}")


def deism_method(json_file_path=None):

    print("deism_method: starting simulation")
    st = time.time()  # start time of calculation
    result_container = {}
    import os

    result_container = {}
    if json_file_path is not None:
        # Step 1: read JSON to get geo_path
        with open(json_file_path, "r") as json_file:
            result_container = json.load(json_file)
            geo_path = os.path.join(
                os.path.dirname(json_file_path), result_container["geo_path"]
            )

        # Step 2: update areas, wall centers, volume
        update_surface_areas(json_file_path, geo_path)
        update_wall_centers(json_file_path, geo_path)
        Volume, room = get_room_geometry(geo_file=geo_path)

        # Step 3: write volume to JSON
        with open(json_file_path, "r+") as json_file:
            data = json.load(json_file)
            data["geometry"][0]["room_volumn"] = Volume
            json_file.seek(0)
            json.dump(data, json_file, indent=4)
            json_file.truncate()

        # Step 4: re-load updated JSON into result_container
        with open(json_file_path, "r") as json_file:
            result_container = json.load(json_file)

    # Checking whether the 'should_cancel' flag has been set to True by the user
    # Do not call this function all the time, as it is quite heavy
    # This function should be called in the main calculation loop
    def check_should_cancel(json_file_path_in):
        try:
            if json_file_path_in is not None:
                with open(json_file_path_in, "r") as json_file_to_check:
                    data = json.load(json_file_to_check)

            # Update the specified field value
            if "should_cancel" in data:
                return data["should_cancel"]

        except Exception as e:
            print("check_should_cancel returned: " + str(e))
            print(traceback.format_exc())

    if check_should_cancel(json_file_path):
        return

    # Load from the json file
    print("Obtaining simulation settings from the json file ... \n")
    if result_container:
        simulation_settings = result_container["simulationSettings"]
        coord_source = [
            result_container["results"][0]["sourceX"],
            result_container["results"][0]["sourceY"],
            result_container["results"][0]["sourceZ"],
        ]

        coord_rec = [
            result_container["results"][0]["responses"][0]["x"],
            result_container["results"][0]["responses"][0]["y"],
            result_container["results"][0]["responses"][0]["z"],
        ]
        abs_coeffs_loaded = result_container["absorption_coefficients"]
        freq_bands = np.array(result_container["results"][0]["frequencies"])

    # Convert data to the ones needed in DEISM
    print("Converting data to the ones needed in DEISM ... \n")
    # -----------------------------------------------------------
    # About room geometry and wall properties
    # N is the number of vertices of the room
    # M is the number of wall centers
    # -----------------------------------------------------------
    vertices = np.array(result_container["geometry"][0]["vertices"])  # Nx3 numpy array
    wall_centers_loaded = result_container["geometry"][0]["wall_centers"]
    room_volumn = result_container["geometry"][0]["room_volumn"]  # float
    room_areas_loaded = result_container["geometry"][0][
        "room_areas"
    ]  # (M,) numpy array
    # we want the absorption has size 6 * len(frequency bands)
    # The first dimension is for the walls, viz., x1, x2, y1, y2, z1, z2
    # Corresponding to wall 1, wall 3, wall 2, wall 4, floor, ceiling
    wall_order = ["wall1", "wall3", "wall2", "wall4", "floor", "ceiling"]
    # Create an empty array for the absorption coefficients
    absorption_coefficients = np.zeros((6, len(freq_bands)))
    wall_centers = np.zeros((6, 3))
    room_areas = np.zeros((6, 1))
    for wall in wall_order:
        absorption_coefficients[wall_order.index(wall), :] = parse_value(
            abs_coeffs_loaded[wall]
        )
        wall_centers[wall_order.index(wall), :] = parse_value(wall_centers_loaded[wall])
        room_areas[wall_order.index(wall), :] = parse_value(room_areas_loaded[wall])

    # Apply DEISM
    deism = DEISM("RIR", room)
    print("valuess of deism")  #
    print("vertices", vertices)
    print("wall center", wall_centers)
    print("room areas", room_areas)

    deism.update_room(
        vertices,
        wall_centers,
        room_volumn,
        room_areas,
    )
    deism.update_wall_materials(
        absorption_coefficients, freq_bands, "absorpCoefficient"
    )
    deism.update_freqs()
    deism.update_source_receiver()
    deism.update_directivities()
    pressure = deism.run_DEISM()
    # -----------------------------------------------------------
    # Save the simulation results
    # -----------------------------------------------------------
    # Save the simulation results in the json file
    result_container["results"][0]["responses"][0][
        "receiverResults"
    ] = pressure.tolist()
    with open(json_file_path, "w") as new_result_json:
        new_result_json.write(json.dumps(result_container, indent=4))
    print("desim_method: simulation done!")


def main():
    deism_method(json_file_path="examples/exampleInput_Deism.json")


# -------------------------------------------------------
if __name__ == "__main__":
    main()



"""
Shared utilities for DEISM modules to avoid circular imports
"""

import numpy as np
from scipy import special as scy


def rotation_matrix_ZXZ(alpha, beta, gamma):
    """
    The rotation matrix calculation used in COMSOL, see:
    https://doc.comsol.com/5.5/doc/com.comsol.help.comsol/comsol_ref_definitions.12.092.html
    """
    a11 = np.cos(alpha) * np.cos(gamma) - np.sin(alpha) * np.cos(beta) * np.sin(gamma)
    a12 = -np.cos(alpha) * np.sin(gamma) - np.sin(alpha) * np.cos(beta) * np.cos(gamma)
    a13 = np.sin(beta) * np.sin(alpha)
    a21 = np.sin(alpha) * np.cos(gamma) + np.cos(alpha) * np.cos(beta) * np.sin(gamma)
    a22 = -np.sin(alpha) * np.sin(gamma) + np.cos(alpha) * np.cos(beta) * np.cos(gamma)
    a23 = -np.sin(beta) * np.cos(alpha)
    a31 = np.sin(beta) * np.sin(gamma)
    a32 = np.sin(beta) * np.cos(gamma)
    a33 = np.cos(beta)
    R = np.array([[a11, a12, a13], [a21, a22, a23], [a31, a32, a33]])
    return R


def SHCs_from_pressure_LS(Psh, Dir_all, sph_order_FEM, freqs_all):
    """
    Obtaining spherical harmonic coefficients using least-square solution
    Input:
    Psh - sampled sound field, size (# frequencies, # samples)
    Dir_all - Directions of the sampling points [azimuth(0-2pi),inclination(0-pi)] in each row
    sph_order_FEM - max. spherical harmonic order
    freqs_all - frequency vector
    Output:
    Pmnr0 - spherical harmonic coefficients, size (# frequencies, # SH. orders, # SH. modes)
    """

    Y = np.zeros([len(Dir_all), (sph_order_FEM + 1) ** 2], dtype=complex)
    for n in range(sph_order_FEM + 1):
        for m in range(-n, n + 1):
            Y[:, n**2 + n + m] = scy.sph_harm(m, n, Dir_all[:, 0], Dir_all[:, 1])
    Y_pinv = np.linalg.pinv(Y)
    fnm = Y_pinv @ Psh.T

    # Convert to the same shape as used in Pmnr0
    Pmnr0 = np.zeros(
        [freqs_all.size, sph_order_FEM + 1, 2 * sph_order_FEM + 1],
        dtype="complex",
    )
    for n in range(sph_order_FEM + 1):
        for m in range(-n, n + 1):
            Pmnr0[:, n, m + n] = fnm[n**2 + n + m, :]

    return Pmnr0



"""
Testing DEISM-LC matrix calculation
"""

# -------------------------------------------------------
# Authors: Zeyu Xu
# Email: zeyu.xu@audiolabs-erlangen.de
# -------------------------------------------------------
import time
import numpy as np
import matplotlib.pyplot as plt
from scipy import special as scy
import ray
import psutil
from deism.core_deism import *


def initialize_params():
    """Initialize parameters related to the simulation."""
    params = {
        # ---------- physical parameters ----------------
        "soundSpeed": 343,  # speed of sound
        "airDensity": 1.2,  # constant of air
        # ---------- room parameters ----------------
        "roomSize": np.array([4, 3, 2.5]),  # Room size, can be randomized a bit
        "maxReflOrder": 10,  # Maximum reflection order
        "acousImpend": 18,  # The acoustic impedance of the walls, can be randomized a bit
        "angDepFlag": 1,  # 1: angle-dependent reflection coefficient, 0: angle-independent reflection coefficient
        # ---------- simulation parameter ----------------
        "sampleRate": 2000,  # Sampling rate
        "RIRLength": 0.3,  # RIR length in seconds
        "startFreq": 20,  # start frequency
        "freqStep": 2,  # frequency step size
        "endFreq": 1000,  # stop frequency
        "posSource": np.array(
            [1.1, 1.1, 1.3]
        ),  # source position, can be randomized a bit
        "posReceiver": np.array(
            [2.9, 1.9, 1.3]
        ),  # receiver position, can be randomized a bit
        "orientSource": np.array(
            [0, 0, 0]
        ),  # Euler angles of the source facing direction
        "orientReceiver": np.array(
            [180, 0, 0]
        ),  # Euler angles of the receiver facing direction
        # ---------- source and receiver parameters ----------------
        # source and receiver parameters, which might not be necessary for version 1.1
        "qFlowStrength": 0.001,  # Point source flow strength used in FEM simulation !!!
        "ifReceiverNormalize": 1,  # If normalize the receiver directivity
        # ---------- DEISM parameters ----------------
        "sourceType": "speaker_cuboid_cyldriver_1",  # Try to use smarter way to load data !!!
        "receiverType": "speaker_cuboid_cyldriver_1",  # Try to use smarter way to load data !!!
        # "num_samples": 1764,  # Number of sample points on the transparent sphere !!!
        # "sampling_scheme": "uniform",  # !!!
        # source and receiver directivity parameters, should change based on loaded data
        "nSourceOrder": 5,  # max. spherical harmonic directivity order
        "radiusSource": np.array([0.4]),  # Radius of transparent sphere
        "vReceiverOrder": 5,  # max. spherical harmonic directivity order
        "radiusReceiver": np.array([0.5]),  # Radius of transparent sphere
        "srcOrient": np.array([0, 0, 0]),  # Euler angles of the source facing direction
        "recOrient": np.array(
            [180, 0, 0]
        ),  # Euler angles of the receiver facing direction
        "ifRemoveDirectPath": 0,
        "silentMode": 0,
    }
    # Compute the rest of the parameters
    params = compute_rest_params(params)
    return params


def main():
    params = initialize_params()
    # Detect conflicts between the parameters
    detect_conflicts(params)
    # Get directivity data for the source and receiver
    params = init_source_directivities(params)
    params = init_receiver_directivities(params)
    # load_directivity_data(params, src_facing, rec_facing)

    # Precompute reflection paths
    DEISMMode = "MIX"
    params["mixEarlyOrder"] = 2
    params["numParaImages"] = 50000
    images = pre_calc_images_src_rec(params)
    # if DEISMMode == "ORG" or DEISMMode == "LC":
    #     # Merge images
    # Used for DEISM-ORG and DEISM-LC
    images_merged = merge_images(images)
    # Precompute Wigner 3J matrices, USED for DEISM-ORG and DEISM-MIX
    Wigner = pre_calc_Wigner(params)

    # Vectorize the directivity data, used for DEISM-LC
    params = vectorize_C_nm_s(params)
    params = vectorize_C_vu_r(params)

    # Initialize Ray
    num_cpus = psutil.cpu_count(logical=False)
    if not ray.is_initialized():
        ray.init(num_cpus=num_cpus)
        print("\n")

    # Simulate room transfer functions
    P_DEISM = ray_run_DEISM(params, images_merged, Wigner)
    # P_DEISM_LC = ray_run_DEISM_LC(params, images_merged)
    P_DEISM_LC_matrix = ray_run_DEISM_LC_matrix(params, images_merged)
    P_DEISM_MIX = ray_run_DEISM_MIX(params, images, Wigner)

    # Shutdown Ray
    ray.shutdown()

    # Load FEM solutions
    freqs_COMSOL, P_COMSOL, mic_pos = load_RTF_data(
        params["silentMode"], "room_two_speaker_cuboid_cyldriver_1"
    )

    # Plot the results
    PLOT_SCALE = "dB"
    IF_FREQS_DB = 0
    IF_SAME_MAGSCALE = 0
    IF_UNWRAP_PHASE = 0
    IF_SAVE_PLOT = 0
    figure_name = "tests"
    save_path = "./outputs/figures"
    P_all = [P_DEISM, P_DEISM_LC_matrix, P_DEISM_MIX, P_COMSOL]
    P_labels = ["DEISM", "DEISM-LC-matrix", "DEISM-MIX", "FEM"]
    P_freqs = [params["freqs"], params["freqs"], params["freqs"], params["freqs"]]
    plot_RTFs(
        figure_name,
        save_path,
        P_all,
        P_labels,
        P_freqs,
        PLOT_SCALE,
        IF_FREQS_DB,
        IF_SAME_MAGSCALE,
        IF_UNWRAP_PHASE,
        IF_SAVE_PLOT,
    )


if __name__ == "__main__":
    main()
